@@ -1,5 +1,5 @@
 package Crypt::Bcrypt;
-$Crypt::Bcrypt::VERSION = '0.007';
+$Crypt::Bcrypt::VERSION = '0.008';
 use strict;
 use warnings;
 
@@ -7,8 +7,9 @@ use XSLoader;
 XSLoader::load('Crypt::Bcrypt');
 
 use Exporter 5.57 'import';
-our @EXPORT_OK = qw(bcrypt bcrypt_check);
+our @EXPORT_OK = qw(bcrypt bcrypt_check bcrypt_hashed bcrypt_check_hashed bcrypt_needs_rehash);
 
+use Digest::SHA 'hmac_sha256';
 use MIME::Base64 2.21 qw(encode_base64);
 
 sub bcrypt {
@@ -21,6 +22,50 @@ sub bcrypt {
 	return _bcrypt_hashpw($password, sprintf '$%s$%02d$%s', $subtype, $cost, $encoded_salt);
 }
 
+my $subtype_qr = qr/2[abxy]/;
+my $cost_qr = qr/\d{2}/;
+my $salt_qr = qr{ [./A-Za-z0-9]{22} }x;
+
+sub bcrypt_hashed {
+	my ($password, $subtype, $cost, $salt, $hash_algorithm) = @_;
+	if ($hash_algorithm) {
+		(my $encoded_salt = encode_base64($salt, "")) =~ tr{A-Za-z0-9+/=}{./A-Za-z0-9}d;
+		my $hashed_password = encode_base64(hmac_sha256($password, $encoded_salt), "");
+		my $hash = bcrypt($hashed_password, $subtype, $cost, $salt);
+		$hash =~ s{ ^ \$ ($subtype_qr) \$ ($cost_qr) \$ ($salt_qr) }{\$bcrypt-sha256\$v=2,t=$1,r=$2\$$3\$}x or die $hash;
+		return $hash;
+	}
+	else {
+		bcrypt($password, $subtype, $cost, $salt);
+	}
+}
+
+sub bcrypt_check_hashed {
+	my ($password, $hash) = @_;
+	if ($hash =~ s/ ^ \$ bcrypt-sha256 \$ v=2,t=($subtype_qr),r=($cost_qr) \$ ($salt_qr) \$ /\$$1\$$2\$$3/x) {
+		return bcrypt_check(encode_base64(hmac_sha256($password, $3), ""), $hash);
+	}
+	else {
+		bcrypt_check($password, $hash);
+	}
+}
+
+sub _get_parameters {
+	my ($hash) = @_;
+	if ($hash =~ / \A \$ ($subtype_qr) \$ ($cost_qr) \$ /x) {
+		return ($1, $2, '');
+	}
+	elsif ($hash =~ / ^ \$ bcrypt-sha256 \$ v=2,t=($subtype_qr),r=($cost_qr) \$ /x) {
+		return ($1, $2, 'sha256');
+	}
+}
+
+sub bcrypt_needs_rehash {
+	my ($hash, $wanted_subtype, $wanted_cost, $wanted_hash) = @_;
+	my ($my_subtype, $my_cost, $my_hash) = _get_parameters($hash);
+	return $my_subtype ne $wanted_subtype || $my_cost < $wanted_cost || $my_hash ne ($wanted_hash || '');
+}
+
 1;
 
 # ABSTRACT: A modern bcrypt implementation
@@ -37,7 +82,7 @@ Crypt::Bcrypt - A modern bcrypt implementation
 
 =head1 VERSION
 
-version 0.007
+version 0.008
 
 =head1 SYNOPSIS
 
@@ -91,13 +136,25 @@ C<$cost> must be between 5 and 31 (inclusive). C<$salt> must be exactly 16 bytes
 
 This checks if the C<$password> satisfies the C<$hash>, and does so in a timing-safe manner.
 
+=head2 bcrypt_hashed($password, $subtype, $cost, $salt, $hash_algorithm)
+
+This works like the C<bcrypt> functions, but pre-hashes the password using the specified hash. This is mainly useful to get around the 72 character limit. Currently only sha256 is supported. This uses a salt-keyed hash to prevent password shucking.
+
+=head2 bcrypt_check_hashed($password, $hash)
+
+This verifies pre-hashed passwords as generated by C<bcrypt_hashed>.
+
+=head2 bcrypt_needs_rehash($hash, $wanted_subtype, $wanted_cost, $wanted_hash = '')
+
+This returns true if the bcrypt hash uses a different subtype, cost or hash algorithm than desired.
+
 =head1 SEE OTHER
 
 =over 4
 
 =item * L<Crypt::Passphrase|Crypt::Passphrase>
 
-This is usually a better approach to managing your passwords, it can use this module via L<Crypt::Passphrase::Bcrypt|Crypt::Passphrase::Bcrypt>. It adds support for automatic pre-hashing, and facilitates upgrading the algorithm parameters or even the algorithm itself.
+This is usually a better approach to managing your passwords, it can use this module via L<Crypt::Passphrase::Bcrypt|Crypt::Passphrase::Bcrypt>. It facilitates upgrading the algorithm parameters or even the algorithm itself.
 
 =item * L<Crypt::Eksblowfish::Bcrypt|Crypt::Eksblowfish::Bcrypt>
 
